@@ -21,13 +21,14 @@ __metaclass__ = type
 
 import json
 import os
+import stat
 import tempfile
 
-from ansible import constants as C
+from ansible.constants import mk_boolean as boolean
+from ansible.errors import AnsibleError, AnsibleFileNotFound
+from ansible.module_utils._text import to_bytes, to_native, to_text
 from ansible.plugins.action import ActionBase
-from ansible.utils.boolean import boolean
 from ansible.utils.hashing import checksum
-from ansible.utils.unicode import to_bytes
 
 
 class ActionModule(ActionBase):
@@ -44,21 +45,20 @@ class ActionModule(ActionBase):
         dest    = self._task.args.get('dest', None)
         raw     = boolean(self._task.args.get('raw', 'no'))
         force   = boolean(self._task.args.get('force', 'yes'))
-        faf     = self._task.first_available_file
         remote_src = boolean(self._task.args.get('remote_src', False))
         follow  = boolean(self._task.args.get('follow', False))
 
-        if (source is None and content is None and faf is None) or dest is None:
-            result['failed'] = True
+        result['failed'] = True
+        if (source is None and content is None) or dest is None:
             result['msg'] = "src (or content) and dest are required"
-            return result
-        elif (source is not None or faf is not None) and content is not None:
-            result['failed'] = True
+        elif source is not None and content is not None:
             result['msg'] = "src and content are mutually exclusive"
-            return result
         elif content is not None and dest is not None and dest.endswith("/"):
-            result['failed'] = True
             result['msg'] = "dest must be a file if content is defined"
+        else:
+            del result['failed']
+
+        if result.get('failed'):
             return result
 
         # Check if the source ends with a "/"
@@ -81,33 +81,27 @@ class ActionModule(ActionBase):
                 source = content_tempfile
             except Exception as err:
                 result['failed'] = True
-                result['msg'] = "could not write content temp file: %s" % err
+                result['msg'] = "could not write content temp file: %s" % to_native(err)
                 return result
 
         # if we have first_available_file in our vars
         # look up the files and use the first one we find as src
-        elif faf:
-            source = self._get_first_available_file(faf, task_vars.get('_original_file', None))
-            if source is None:
-                result['failed'] = True
-                result['msg'] = "could not find src in first_available_file list"
-                return result
-
         elif remote_src:
-            result.update(self._execute_module(module_name='copy', module_args=self._task.args, task_vars=task_vars, delete_remote_tmp=False))
+            result.update(self._execute_module(task_vars=task_vars))
             return result
-
-        else:
-            if self._task._role is not None:
-                source = self._loader.path_dwim_relative(self._task._role._role_path, 'files', source)
-            else:
-                source = self._loader.path_dwim_relative(self._loader.get_basedir(), 'files', source)
+        else:  # find in expected paths
+            try:
+                source = self._find_needle('files', source)
+            except AnsibleError as e:
+                result['failed'] = True
+                result['msg'] = to_text(e)
+                return result
 
         # A list of source file tuples (full_path, relative_path) which will try to copy to the destination
         source_files = []
 
         # If source is a directory populate our list else source is a file and translate it to a tuple.
-        if os.path.isdir(to_bytes(source, errors='strict')):
+        if os.path.isdir(to_bytes(source, errors='surrogate_or_strict')):
             # Get the amount of spaces to remove to get the relative path.
             if source_trailing_slash:
                 sz = len(source)
@@ -115,14 +109,18 @@ class ActionModule(ActionBase):
                 sz = len(source.rsplit('/', 1)[0]) + 1
 
             # Walk the directory and append the file tuples to source_files.
-            for base_path, sub_folders, files in os.walk(source):
+            for base_path, sub_folders, files in os.walk(to_bytes(source)):
                 for file in files:
-                    full_path = os.path.join(base_path, file)
+                    full_path = to_text(os.path.join(base_path, file), errors='surrogate_or_strict')
                     rel_path = full_path[sz:]
                     if rel_path.startswith('/'):
                         rel_path = rel_path[1:]
                     source_files.append((full_path, rel_path))
 
+                # recurse into subdirs
+                for sf in sub_folders:
+                    source_files += self._get_recursive_files(os.path.join(source, to_text(sf)), sz=sz)
+
             # If it's recursive copy, destination is always a dir,
             # explicitly mark it so (note - copy module relies on this).
             if not self._connection._shell.path_has_trailing_slash(dest):
@@ -141,30 +139,34 @@ class ActionModule(ActionBase):
         delete_remote_tmp = (len(source_files) == 1)
 
         # If this is a recursive action create a tmp path that we can share as the _exec_module create is too late.
-        remote_user = task_vars.get('ansible_ssh_user') or self._play_context.remote_user
         if not delete_remote_tmp:
             if tmp is None or "-tmp-" not in tmp:
-                tmp = self._make_tmp_path(remote_user)
-                self._cleanup_remote_tmp = True
+                tmp = self._make_tmp_path()
 
         # expand any user home dir specifier
         dest = self._remote_expand_user(dest)
 
+        # Keep original value for mode parameter
+        mode_value = self._task.args.get('mode', None)
+
         diffs = []
         for source_full, source_rel in source_files:
 
-            source_full = self._loader.get_real_file(source_full)
-
-            # Generate a hash of the local file.
-            local_checksum = checksum(source_full)
-
-            # If local_checksum is not defined we can't find the file so we should fail out.
-            if local_checksum is None:
+            # If the local file does not exist, get_real_file() raises AnsibleFileNotFound
+            try:
+                source_full = self._loader.get_real_file(source_full)
+            except AnsibleFileNotFound as e:
                 result['failed'] = True
-                result['msg'] = "could not find src=%s" % source_full
+                result['msg'] = "could not find src=%s, %s" % (source_full, e)
                 self._remove_tmp_path(tmp)
                 return result
 
+            # Get the local mode and set if user wanted it preserved
+            # https://github.com/ansible/ansible-modules-core/issues/1124
+            if self._task.args.get('mode', None) == 'preserve':
+                lmode = '0%03o' % stat.S_IMODE(os.stat(source_full).st_mode)
+                self._task.args['mode'] = lmode
+
             # This is kind of optimization - if user told us destination is
             # dir, do path manipulation right away, otherwise we still check
             # for dest being a dir via remote call below.
@@ -174,7 +176,7 @@ class ActionModule(ActionBase):
                 dest_file = self._connection._shell.join_path(dest)
 
             # Attempt to get remote file info
-            dest_status = self._execute_remote_stat(dest_file, all_vars=task_vars, follow=follow, tmp=tmp)
+            dest_status = self._execute_remote_stat(dest_file, all_vars=task_vars, follow=follow, tmp=tmp, checksum=force)
 
             if dest_status['exists'] and dest_status['isdir']:
                 # The dest is a directory.
@@ -188,12 +190,15 @@ class ActionModule(ActionBase):
                 else:
                     # Append the relative source location to the destination and get remote stats again
                     dest_file = self._connection._shell.join_path(dest, source_rel)
-                    dest_status = self._execute_remote_stat(dest_file, all_vars=task_vars, follow=follow, tmp=tmp)
+                    dest_status = self._execute_remote_stat(dest_file, all_vars=task_vars, follow=follow, tmp=tmp, checksum=force)
 
             if dest_status['exists'] and not force:
-                # remote_file does not exist so continue to next iteration.
+                # remote_file exists so continue to next iteration.
                 continue
 
+            # Generate a hash of the local file.
+            local_checksum = checksum(source_full)
+
             if local_checksum != dest_status['checksum']:
                 # The checksums don't match and we will change or error out.
                 changed = True
@@ -202,8 +207,7 @@ class ActionModule(ActionBase):
                 # If this is recursive we already have a tmp path.
                 if delete_remote_tmp:
                     if tmp is None or "-tmp-" not in tmp:
-                        tmp = self._make_tmp_path(remote_user)
-                        self._cleanup_remote_tmp = True
+                        tmp = self._make_tmp_path()
 
                 if self._play_context.diff and not raw:
                     diffs.append(self._get_diff_data(dest_file, source_full, task_vars))
@@ -217,8 +221,10 @@ class ActionModule(ActionBase):
                 # Define a remote directory that we will copy the file to.
                 tmp_src = self._connection._shell.join_path(tmp, 'source')
 
+                remote_path = None
+
                 if not raw:
-                    self._transfer_file(source_full, tmp_src)
+                    remote_path = self._transfer_file(source_full, tmp_src)
                 else:
                     self._transfer_file(source_full, dest_file)
 
@@ -227,7 +233,8 @@ class ActionModule(ActionBase):
                 self._loader.cleanup_tmp_file(source_full)
 
                 # fix file permissions when the copy is done as a different user
-                self._fixup_perms(tmp, remote_user, recursive=True)
+                if remote_path:
+                    self._fixup_perms2((tmp, remote_path))
 
                 if raw:
                     # Continue to next iteration if raw is defined.
@@ -245,8 +252,12 @@ class ActionModule(ActionBase):
                         original_basename=source_rel,
                     )
                 )
+                if 'content' in new_module_args:
+                    del new_module_args['content']
 
-                module_return = self._execute_module(module_name='copy', module_args=new_module_args, task_vars=task_vars, tmp=tmp, delete_remote_tmp=delete_remote_tmp)
+                module_return = self._execute_module(module_name='copy',
+                        module_args=new_module_args, task_vars=task_vars,
+                        tmp=tmp, delete_remote_tmp=delete_remote_tmp)
                 module_executed = True
 
             else:
@@ -260,6 +271,14 @@ class ActionModule(ActionBase):
                     self._remove_tmp_path(tmp)
                     continue
 
+                # Fix for https://github.com/ansible/ansible-modules-core/issues/1568.
+                # If checksums match, and follow = True, find out if 'dest' is a link. If so,
+                # change it to point to the source of the link.
+                if follow:
+                    dest_status_nofollow = self._execute_remote_stat(dest_file, all_vars=task_vars, follow=False)
+                    if dest_status_nofollow['islnk'] and 'lnk_source' in dest_status_nofollow.keys():
+                        dest = dest_status_nofollow['lnk_source']
+
                 # Build temporary module_args.
                 new_module_args = self._task.args.copy()
                 new_module_args.update(
@@ -271,7 +290,9 @@ class ActionModule(ActionBase):
                 )
 
                 # Execute the file module.
-                module_return = self._execute_module(module_name='file', module_args=new_module_args, task_vars=task_vars, tmp=tmp, delete_remote_tmp=delete_remote_tmp)
+                module_return = self._execute_module(module_name='file',
+                        module_args=new_module_args, task_vars=task_vars,
+                        tmp=tmp, delete_remote_tmp=delete_remote_tmp)
                 module_executed = True
 
             if not module_return.get('checksum'):
@@ -289,6 +310,9 @@ class ActionModule(ActionBase):
             if 'path' in module_return and 'dest' not in module_return:
                 module_return['dest'] = module_return['path']
 
+            # reset the mode
+            self._task.args['mode'] = mode_value
+
         # Delete tmp path if we were recursive or if we did not execute a module.
         if not delete_remote_tmp or (delete_remote_tmp and not module_executed):
             self._remove_tmp_path(tmp)
@@ -303,6 +327,22 @@ class ActionModule(ActionBase):
 
         return result
 
+    def _get_recursive_files(self, topdir, sz=0):
+        ''' Recursively create file tuples for sub folders '''
+        r_files = []
+        for base_path, sub_folders, files in os.walk(to_bytes(topdir)):
+            for fname in files:
+                full_path = to_text(os.path.join(base_path, fname), errors='surrogate_or_strict')
+                rel_path = full_path[sz:]
+                if rel_path.startswith('/'):
+                    rel_path = rel_path[1:]
+                r_files.append((full_path, rel_path))
+
+                for sf in sub_folders:
+                    r_files += self._get_recursive_files(os.path.join(topdir, to_text(sf)), sz=sz)
+
+        return r_files
+
     def _create_content_tempfile(self, content):
         ''' Create a tempfile containing defined content '''
         fd, content_tempfile = tempfile.mkstemp()
@@ -310,8 +350,6 @@ class ActionModule(ActionBase):
         content = to_bytes(content)
         try:
             f.write(content)
-            if not content.endswith('\n'):
-              f.write('\n')
         except Exception as err:
             os.remove(content_tempfile)
             raise Exception(err)
